@@ -11,7 +11,9 @@ open Ctypes_static
 open Ctypes_path
 open Cstubs_errors
 
-type concurrency_policy = [ `Sequential | `Lwt_jobs ]
+type non_lwt = [ `Sequential | `Unlocked ]
+type lwt = [ `Lwt_jobs | `Lwt_preemptive ]
+type concurrency_policy = [ non_lwt | lwt ]
 type errno_policy = [ `Ignore_errno | `Return_errno ]
 
 type lident = string
@@ -23,7 +25,7 @@ type ml_type = [ `Ident of path
 type ml_external_type = [ `Prim of ml_type list * ml_type ]
 
 type ml_pat = [ `Var of string
-              | `Record of (path * ml_pat) list
+              | `Record of (path * ml_pat) list * [`Etc | `Complete]
               | `As of ml_pat * string
               | `Underscore
               | `Con of path * ml_pat list ]
@@ -170,7 +172,8 @@ struct
   let rec ml_pat appl_parens fmt pat =
     match appl_parens, pat with
     | _, `Var x -> fprintf fmt "%s" x
-    | _, `Record fs -> fprintf fmt "{@[%a}@]" pat_fields fs
+    | _, `Record (fs, `Etc) -> fprintf fmt "{@[%a_}@]" pat_fields fs
+    | _, `Record (fs, `Complete) -> fprintf fmt "{@[%a}@]" pat_fields fs
     | _, `As (p, x) -> fprintf fmt "@[(%a@ as@ %s)@]" (ml_pat NoApplParens) p x
     | _, `Underscore -> fprintf fmt "_"
     | _, `Con (c, []) -> fprintf fmt "%a" format_path c
@@ -289,9 +292,9 @@ let rec ml_external_type_of_fn :
   type a. concurrency:concurrency_policy -> errno:errno_policy ->
   a fn -> polarity -> ml_external_type =
   fun ~concurrency ~errno fn polarity -> match fn, concurrency, errno with
-    | Returns t, `Sequential, `Ignore_errno ->
+    | Returns t, (#non_lwt|`Lwt_preemptive), `Ignore_errno ->
       `Prim ([], ml_typ_of_typ polarity t)
-    | Returns t, `Sequential, `Return_errno ->
+    | Returns t, (#non_lwt|`Lwt_preemptive), `Return_errno ->
       `Prim ([], `Pair (ml_typ_of_typ polarity t, int_type))
     | Returns t, `Lwt_jobs, `Ignore_errno ->
       `Prim ([], `Appl (lwt_job_type, [ml_typ_of_typ polarity t]))
@@ -330,13 +333,13 @@ let make_structured = Ctypes_path.path_of_string "CI.make_structured"
 let map_result ~concurrency ~errno f e =
   let map_result f x = `Appl (`Appl (`Ident map_result_id, f), x) in
   match concurrency, errno, f with
-    `Sequential, `Ignore_errno, `MakePtr x ->
+    #non_lwt, `Ignore_errno, `MakePtr x ->
     `MakePtr (`Ident (path_of_string x), e)
-  | `Sequential, `Ignore_errno, `MakeFunPtr x ->
+  | #non_lwt, `Ignore_errno, `MakeFunPtr x ->
     `MakeFunPtr (`Ident (path_of_string x), e)
-  | `Sequential, `Ignore_errno, `MakeStructured x ->
+  | #non_lwt, `Ignore_errno, `MakeStructured x ->
     `MakeStructured (`Ident (path_of_string x), e)
-  | `Sequential, `Ignore_errno, `Appl x ->
+  | #non_lwt, `Ignore_errno, `Appl x ->
     `Appl (`Ident (path_of_string x), e)
   | _, _, `MakePtr x ->
     map_result (`Appl (`Ident make_ptr, `Ident (path_of_string x))) e
@@ -356,18 +359,24 @@ let rec pattern_and_exp_of_typ : type a. concurrency:concurrency_policy -> errno
     let id = Cstubs_public_name.constructor_cident_of_prim ~module_name:"CI" p in
     (static_con "Primitive" [`Con (id, [])], None, binds)
   | Pointer _ ->
-    let x = fresh_var () in
-    let pat = static_con "Pointer" [`Var x] in
     begin match pol with
-    | In -> (pat, Some (`Appl (`Ident (path_of_string "CI.cptr"), e)), binds)
-    | Out -> (pat, Some (map_result ~concurrency ~errno (`MakePtr x) e), binds)
+    | In ->
+      let pat = static_con "Pointer" [`Underscore] in
+      (pat, Some (`Appl (`Ident (path_of_string "CI.cptr"), e)), binds)
+    | Out ->
+      let x = fresh_var () in
+      let pat = static_con "Pointer" [`Var x] in
+      (pat, Some (map_result ~concurrency ~errno (`MakePtr x) e), binds)
     end
   | Funptr _ ->
-    let x = fresh_var () in
-    let pat = static_con "Funptr" [`Var x] in
     begin match pol with
-    | In -> (pat, Some (`Appl (`Ident (path_of_string "CI.fptr"), e)), binds)
-    | Out -> (pat, Some (map_result ~concurrency ~errno (`MakeFunPtr x) e), binds)
+    | In ->
+      let pat = static_con "Funptr" [`Underscore] in
+      (pat, Some (`Appl (`Ident (path_of_string "CI.fptr"), e)), binds)
+    | Out ->
+      let x = fresh_var () in
+      let pat = static_con "Funptr" [`Var x] in
+      (pat, Some (map_result ~concurrency ~errno (`MakeFunPtr x) e), binds)
     end
   | Struct _ ->
     begin match pol with
@@ -400,16 +409,16 @@ let rec pattern_and_exp_of_typ : type a. concurrency:concurrency_policy -> errno
       let (p, None, binds), e | (p, Some e, binds), _ =
         pattern_and_exp_of_typ ~concurrency ~errno ty e pol binds, e in
       let pat = static_con "View"
-        [`Record [path_of_string "CI.ty", p;
-                  path_of_string "write", `Var x]] in
+        [`Record ([path_of_string "CI.ty", p;
+                   path_of_string "write", `Var x], `Etc)] in
       (pat, Some (`Ident (Ctypes_path.path_of_string y)), (y, e) :: binds)
     | Out ->
       let (p, None, binds), e | (p, Some e, binds), _ =
         pattern_and_exp_of_typ ~concurrency ~errno ty e pol binds, e in
       let x = fresh_var () in
       let pat = static_con "View"
-        [`Record [path_of_string "CI.ty", p;
-                  path_of_string "read", `Var x]] in
+        [`Record ([path_of_string "CI.ty", p;
+                   path_of_string "read", `Var x], `Etc)] in
       (pat, Some (map_result ~concurrency ~errno (`Appl x) e), binds)
     end
   | OCaml ty ->
@@ -450,7 +459,11 @@ let rec pattern_of_typ : type a. a typ -> ml_pat = function
     static_con "Union" [`Underscore]
   | View { ty } ->
     static_con "View"
-      [`Record [path_of_string "CI.ty", pattern_of_typ ty]]
+      [`Record ([path_of_string "CI.ty", pattern_of_typ ty], `Etc)]
+  | Array (_, _) ->
+     static_con "Array" [`Underscore; `Underscore]
+  | Bigarray _ ->
+     static_con "Bigarray" [`Underscore]
   | OCaml String ->
     Ctypes_static.unsupported
       "cstubs does not support OCaml strings as global values"
@@ -464,14 +477,6 @@ let rec pattern_of_typ : type a. a typ -> ml_pat = function
     internal_error
       "Unexpected abstract type encountered during ML code generation: %s"
       (Ctypes.string_of_typ ty)
-  | Array _ as ty ->
-    internal_error
-      "Unexpected array type encountered during ML code generation: %s"
-      (Ctypes.string_of_typ ty)
-  | Bigarray _ as ty ->
-    internal_error
-      "Unexpected bigarray type encountered during ML code generation: %s"
-      (Ctypes.string_of_typ ty)
 
 type wrapper_state = {
   pat: ml_pat;
@@ -482,10 +487,15 @@ type wrapper_state = {
 }
 
 let lwt_unix_run_job = Ctypes_path.path_of_string "Lwt_unix.run_job"
+let lwt_preemptive_detach = Ctypes_path.path_of_string "Lwt_preemptive.detach"
 
 let run_exp ~concurrency exp = match concurrency with
-    `Sequential -> exp
+    #non_lwt -> exp
   | `Lwt_jobs -> `Appl (`Ident lwt_unix_run_job, exp)
+  | `Lwt_preemptive -> `Appl
+                         (`Appl (`Ident lwt_preemptive_detach,
+                                 `Fun (["_"], exp)),
+                          `Unit)
 
 let let_bind : (lident * ml_exp) list -> ml_exp -> ml_exp =
   fun binds e ->
@@ -543,11 +553,11 @@ let wrapper : type a. concurrency:concurrency_policy -> errno:errno_policy ->
   fun ~concurrency ~errno id fn f pol ->
     let p = wrapper_body ~concurrency ~errno fn (`Ident (path_of_string f)) pol [] in
     match p, concurrency with
-      { trivial = true; pat; binds }, `Sequential ->
+      { trivial = true; pat; binds }, #non_lwt ->
       (pat, let_bind binds (run_exp ~concurrency (`Ident id)))
-    | { exp; args; pat; binds }, `Sequential ->
+    | { exp; args; pat; binds }, #non_lwt ->
       (pat, `Fun (args, let_bind binds exp))
-    | { trivial = true; pat; args; binds }, `Lwt_jobs ->
+    | { trivial = true; pat; args; binds }, #lwt ->
       let exp : ml_exp = List.fold_left (fun f p -> `Appl (f, `Ident (path_of_string p))) (`Ident id) args in
       (pat, `Fun (args,
                   let_bind binds
@@ -555,7 +565,7 @@ let wrapper : type a. concurrency:concurrency_policy -> errno:errno_policy ->
                             `Appl (`Appl (`Ident lwt_bind,
                                           run_exp ~concurrency exp),
                                    return_result ~args:(args @ (List.map fst binds)))))))
-    | { exp; args; pat; binds }, `Lwt_jobs ->
+    | { exp; args; pat; binds }, #lwt ->
       (pat, `Fun (args,
                   let_bind binds
                     (`Appl (`Ident box_lwt,
