@@ -12,11 +12,12 @@
 
 #include "ctypes_primitives.h"
 #include "ctypes_complex_stubs.h"
+#include "ctypes_ldouble_stubs.h"
 #include "ctypes_raw_pointer.h"
 #include "ctypes_managed_buffer_stubs.h"
 #include <caml/threads.h>
 #define CTYPES_PTR_OF_OCAML_STRING(s) \
-  (String_val(Field(s, 1)) + Int_val(Field(s, 0)))
+  (String_val(Field(s, 1)) + Long_val(Field(s, 0)))
 #define Ctypes_val_char(c) \
   (Val_int((c + 256) % 256))
 #define CTYPES_PAIR_WITH_ERRNO(v)
@@ -33,4 +34,10 @@ static inline value ctypes_pair_with_errno(value p)
   CAMLreturn (v);
 }
 
+#if defined(__MINGW32__) || defined(__MINGW64__)
+#define ctypes_printf __mingw_printf
+#else
+#define ctypes_printf printf
+#endif
+
 #endif /* CTYPES_CSTUBS_INTERNALS_H */
