@@ -14,14 +14,22 @@
 #include <float.h>
 #include <assert.h>
 #include <string.h>
-#include <complex.h>
+#include <errno.h>
+#include <math.h>
 
 #if defined _WIN32 && !defined __CYGWIN__
 #include <windows.h>
 #else
+#include <pthread.h>
+#include <unistd.h>
+#if defined(__APPLE__)
+#include <dispatch/dispatch.h>
+#else
 #include <semaphore.h>
 #endif
+#endif
 
+#include "ctypes_complex_compatibility.h"
 #include "test_functions.h"
 
 static int add(int x, int y) { return x + y; }
@@ -397,46 +405,96 @@ float retrieve_FLT_MAX(void) { return FLT_MAX; }
 double retrieve_DBL_MIN(void) { return DBL_MIN; }
 double retrieve_DBL_MAX(void) { return DBL_MAX; }
 
-void add_complexd(double complex *l, double complex *r, double complex *out)
+void add_complexd(double _Complex *l, double _Complex *r, double _Complex *out)
 {
   *out = *l + *r;
 }
 
-void mul_complexd(double complex *l, double complex *r, double complex *out)
+void mul_complexd(double _Complex *l, double _Complex *r, double _Complex *out)
 {
   *out = *l * *r;
 }
 
-void add_complexf(float complex *l, float complex *r, float complex *out)
+void rotdist_complexd(double _Complex *c, double *r, double *out) {
+  double _Complex x = *c * (ctypes_compat_make_complex(cos(*r), sin(*r)));
+  *out = fabs(ctypes_compat_creal(x)) + fabs(ctypes_compat_cimag(x));
+}
+
+void add_complexld(long double _Complex *l, long double _Complex *r, long double _Complex *out)
 {
   *out = *l + *r;
 }
 
-void mul_complexf(float complex *l, float complex *r, float complex *out)
+void mul_complexld(long double _Complex *l, long double _Complex *r, long double _Complex *out)
 {
   *out = *l * *r;
 }
 
-double complex add_complexd_val(double complex l, double complex r)
+void rotdist_complexld(long double _Complex *c, long double *r, long double *out) {
+  long double _Complex x = *c * (ctypes_compat_make_complexl(cosl(*r), sinl(*r)));
+  *out = fabsl(ctypes_compat_creall(x)) + fabsl(ctypes_compat_cimagl(x));
+}
+
+void add_complexf(float _Complex *l, float _Complex *r, float _Complex *out)
+{
+  *out = *l + *r;
+}
+
+void mul_complexf(float _Complex *l, float _Complex *r, float _Complex *out)
+{
+  *out = *l * *r;
+}
+
+void rotdist_complexf(float _Complex *c, float *r, float *out) {
+  float _Complex x = *c * (ctypes_compat_make_complexf(cosf(*r), sinf(*r)));
+  *out = fabsf(ctypes_compat_crealf(x)) + fabsf(ctypes_compat_cimagf(x));
+}
+
+long double _Complex add_complexld_val(long double _Complex l, long double _Complex r)
 {
   return l + r;
 }
 
-double complex mul_complexd_val(double complex l, double complex r)
+long double _Complex mul_complexld_val(long double _Complex l, long double _Complex r)
 {
   return l * r;
 }
 
-float complex add_complexf_val(float complex l, float complex r)
+long double rotdist_complexld_val(long double _Complex c, long double r) {
+  long double _Complex x = c * (ctypes_compat_make_complexl(cosl(r), sinl(r)));
+  return fabsl(ctypes_compat_creall(x)) + fabsl(ctypes_compat_cimagl(x));
+}
+
+double _Complex add_complexd_val(double _Complex l, double _Complex r)
 {
   return l + r;
 }
 
-float complex mul_complexf_val(float complex l, float complex r)
+double _Complex mul_complexd_val(double _Complex l, double _Complex r)
 {
   return l * r;
 }
 
+double rotdist_complexd_val(double _Complex c, double r) {
+  double _Complex x = c * (ctypes_compat_make_complex(cos(r), sin(r)));
+  return fabs(ctypes_compat_creal(x)) + fabs(ctypes_compat_cimag(x));
+}
+
+float _Complex add_complexf_val(float _Complex l, float _Complex r)
+{
+  return l + r;
+}
+
+float _Complex mul_complexf_val(float _Complex l, float _Complex r)
+{
+  return l * r;
+}
+
+float rotdist_complexf_val(float _Complex c, float r) {
+  float _Complex x = c * (ctypes_compat_make_complexf(cosf(r), sinf(r)));
+  return fabsf(ctypes_compat_crealf(x)) + fabsf(ctypes_compat_cimagf(x));
+}
+
 static int (*global_stored_callback)(int) = NULL;
 
 void store_callback(int (*callback)(int))
@@ -515,34 +573,62 @@ void call_registered_callback(int times, int starting_value)
   }
 }
 
-#if defined _WIN32 && !defined __CYGWIN__
+#ifdef __APPLE__
+#define sem_t dispatch_semaphore_t
+#define sem_init(sem, sem_attr1, sem_init_value)                \
+  ((*sem = dispatch_semaphore_create(sem_init_value)) == NULL)
+#define sem_wait(sem)                                   \
+  dispatch_semaphore_wait(*sem, DISPATCH_TIME_FOREVER)
+#define sem_post(sem)                           \
+  (dispatch_semaphore_signal(*sem),0)
+#define sem_destroy(sem)                        \
+  (dispatch_release(*sem),0)
+
+#elif defined(_WIN32) && !defined(__CYGWIN__)
 #define sem_t HANDLE
 #define sem_init(sem, sem_attr1, sem_init_value)        \
-  (void)((*sem = CreateSemaphore(NULL,0,32768,NULL))==NULL)
+  ((*sem = CreateSemaphore(NULL,0,32768,NULL)) == NULL)
 #define sem_wait(sem) \
-  (void)(WAIT_OBJECT_0 != WaitForSingleObject(*sem,INFINITE))
-#define sem_post(sem) (void)ReleaseSemaphore(*sem,1,NULL)
+  (WAIT_OBJECT_0 != WaitForSingleObject(*sem,INFINITE))
+#define sem_post(sem) (ReleaseSemaphore(*sem,1,NULL) == 0)
+#define sem_destroy(sem) (CloseHandle(*sem) == 0 )
 #endif
 
 static sem_t semaphore1;
 static sem_t semaphore2;
+static int semaphores_intialized;
 
 void initialize_waiters(void)
 {
-  sem_init(&semaphore1, 0, -1);
-  sem_init(&semaphore2, 0, -1);
+  if ( semaphores_intialized ) {
+    assert ( sem_destroy(&semaphore1) == 0 );
+    assert ( sem_destroy(&semaphore2) == 0 );
+  }
+  assert ( sem_init(&semaphore1, 0, 0) == 0 );
+  assert ( sem_init(&semaphore2, 0, 0) == 0 );
+  semaphores_intialized = 1;
 }
 
 void post1_wait2(void)
 {
-  sem_post(&semaphore1);
-  sem_wait(&semaphore2);
+  int e;
+  assert ( sem_post(&semaphore1) == 0 );
+  errno = 0;
+  do {
+    e = sem_wait(&semaphore2);
+  } while ( e && errno == EINTR );
+  assert ( e == 0 );
 }
 
 void post2_wait1(void)
 {
-  sem_post(&semaphore2);
-  sem_wait(&semaphore1);
+  int e;
+  assert ( sem_post(&semaphore2) == 0 );
+  errno = 0;
+  do {
+    e = sem_wait(&semaphore1);
+  } while ( e && errno == EINTR );
+  assert ( e == 0 );
 }
 
 size_t sizeof_s1(void) { return sizeof(struct s1); }
@@ -638,7 +724,172 @@ int return_10(void)
   return 10;
 }
 
+void return_void(int *x)
+{
+  *x = 10;
+  return;
+}
+
 int callback_returns_char_a(char (*f)(void))
 {
   return f() == 'a' ? 1 : 0;
 }
+
+#define GEN_RETURN_F(type)                          \
+  type callback_returns_ ## type (type  (*f)(void)) \
+  {                                                 \
+    type x = f();                                   \
+    return x;                                       \
+  }                                                 \
+
+GEN_RETURN_F(uint8_t)
+GEN_RETURN_F(uint16_t)
+GEN_RETURN_F(uint32_t)
+GEN_RETURN_F(uint64_t)
+
+GEN_RETURN_F(int8_t)
+GEN_RETURN_F(int16_t)
+GEN_RETURN_F(int32_t)
+GEN_RETURN_F(int64_t)
+
+GEN_RETURN_F(float)
+GEN_RETURN_F(double)
+GEN_RETURN_F(bool)
+
+char *string_array[2] = { "Hello", "world" };
+int32_t int_array[5] = { 0, 1, 2, 3, 4 };
+
+void check_ones(const int *p, size_t sz)
+{
+  unsigned i = 0;
+  for (; i < sz; i++) {
+    assert (p[i] == 1);
+  }
+}
+
+intnat max_caml_int(void)
+{
+  return (intnat)(((uintnat)-1) / 4);
+}
+
+static
+uint64_t thread_id(void)
+{
+#ifdef _WIN32
+  return (GetCurrentThreadId());
+#else
+  /* if pthread_t is a struct greater than uint64_t,
+     the test could fail ... */
+  union {
+      uint64_t i;
+      pthread_t t;
+  } u;
+  memset(&u, 0, sizeof(u));
+  u.t = pthread_self();
+  return u.i;
+#endif
+}
+
+#ifndef _WIN32
+typedef pthread_t thread_t;
+typedef void *(*start_routine)(void *);
+#else
+typedef HANDLE thread_t;
+typedef DWORD WINAPI (*start_routine)(void *);
+#endif
+
+static int thread_create(thread_t *t, start_routine f, void * param)
+{
+#ifndef _WIN32
+  return (pthread_create(t, NULL, f, param));
+#else
+  HANDLE h = CreateThread(NULL,0,f,param,0,NULL);
+  *t = h;
+  return ( h == NULL );
+#endif
+}
+
+static int thread_join(thread_t t)
+{
+#ifndef _WIN32
+  return ( pthread_join(t, NULL) );
+#else
+  WaitForSingleObject(t, INFINITE);
+  return 0;
+#endif
+}
+
+typedef struct {
+    void (*f)(uint64_t);
+    const unsigned n_callback;
+} t_info;
+
+static void call_multiple_times_r(void *fp)
+{
+  const t_info * t = fp;
+  void (*f)(uint64_t) = t->f;
+  const unsigned n_callback = t->n_callback;
+  const uint64_t tid = thread_id();
+  unsigned i;
+  for ( i = 0 ; i < n_callback ; ++i ) {
+    f(tid);
+  }
+}
+
+#ifndef _WIN32
+static void *call_multiple_times(void *fp)
+{
+  call_multiple_times_r(fp);
+  return NULL;
+}
+#else
+static DWORD WINAPI call_multiple_times(void *fp)
+{
+  call_multiple_times_r(fp);
+  return 0;
+}
+#endif
+
+int foreign_thread_registration_test(void (*test_f)(uint64_t),
+                                     unsigned n_threads,
+                                     unsigned n_callback)
+{
+  thread_t * h_thread;
+  unsigned i;
+  int ret_code = 0;
+  unsigned i_max = 0;
+  t_info thread_info = { .f = test_f , .n_callback = n_callback };
+  const uint64_t tid = thread_id();
+
+  h_thread = malloc(n_threads * (sizeof *h_thread));
+  if ( h_thread == NULL ){
+    fputs("malloc failed\n",stderr);
+    return 1;
+  }
+
+  for ( i = 0 ; i < n_threads ; ++i ) {
+    if ( thread_create(&h_thread[i], call_multiple_times, &thread_info) ) {
+      fputs("Error creating thread\n",stderr);
+      ret_code = 1;
+      break;
+    }
+  }
+  i_max = i;
+
+  for ( i = 0 ; i < i_max ; ++i ) {
+    if ( i < n_callback ) {
+      test_f(tid);
+    }
+    if ( thread_join(h_thread[i]) ) {
+      fputs("Error joining thread\n",stderr);
+      ret_code = 1;
+    }
+  }
+
+  for ( i = n_threads ; i < n_callback ; ++i ) {
+    test_f(tid);
+  }
+
+  free(h_thread);
+  return ret_code;
+}
