@@ -9,7 +9,7 @@
 
 open Ctypes
 
-module Common (F: Cstubs.FOREIGN) =
+module Common (F: Ctypes.FOREIGN) =
 struct
   let s : [`global_struct] structure typ = structure "global_struct"
   let (-:) ty label = field s label ty
@@ -25,10 +25,13 @@ struct
 
   let sum = F.(foreign "sum_range_with_plus_callback"
                  (int @-> int @-> returning int))
+
+  let string_array = F.(foreign_value "string_array" (array 2 string))
+  let int_array = F.(foreign_value "int_array" (bigarray array1 5 Bigarray.int32))
 end
 
 
-module Stubs (F: Cstubs.FOREIGN) =
+module Stubs (F: Ctypes.FOREIGN) =
 struct
   include Common(F)
   let environ = F.(foreign_value "environ" (ptr string_opt))
